::cisco::eem::event_register_cli pattern ".* \\\| " default 60 maxrun 60 questionmark tab enter
#-
# ###########################################################################
# Copyright (c) 2010 Mike Timm <mtimm@cisco.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of Cisco, the name of the copyright holder nor the
#    names of their respective contributors may be used to endorse or
#    promote products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#
# TECHNICAL ASSISTANCE CENTER (TAC) SUPPORT IS NOT AVAILABLE FOR THIS SCRIPT.
# For questions or help, send email to mtimm@cisco.com.
#
# ###########################################################################
# This Tcl policy adds the output modifier 'sort' to IOS commands that
# support output modifiers - mostly show commands.  In 15.0 on a 2800 the
# following output modifiers are supported:
#
# Router#show ver | ?   
#  append    Append redirected output to URL (URLs supporting append operation
#            only)
#  begin     Begin with the line that matches
#  exclude   Exclude lines that match
#  format    Format the output using the specified spec file
#  include   Include lines that match
#  redirect  Redirect output to URL
#  section   Filter a section of output
#  tee       Copy output to URL
#
# When this policy is registered the sort output modifier becomes available:
#
# Router#show ver | ?   
#  append    Append redirected output to URL (URLs supporting append operation
#            only)
#  begin     Begin with the line that matches
#  exclude   Exclude lines that match
#  format    Format the output using the specified spec file
#  include   Include lines that match
#  redirect  Redirect output to URL
#  section   Filter a section of output
#  sort      Sort by column (Added by EEM)
#  tee       Copy output to URL
#
# The sort output modifier requires one option to specify the column number
# which the sort will be done over:
#
# Router#show ver | so ?
#  <0-1000>  Column number
#
# Router#show ver | so 1 ?
#  <cr>
#
# The sort output modifier sorts based on column number throwing away any 
# line that does not contain enough columns and any lines that contain
# non-integer values in the column specified.  While this type of a sort
# is fairly limited in what it can be used for, this is a proof of concept
# that can be extended to do other types of sorts if needed.  In this case
# the sort is a decdending dictionary sort based on an index using Tcl's
# standard lsort.
#
# Example usage, find the largest number of memory blocks allocated by a
# single allocator:
#
#
# Router#show mem proc alloc total | sort 3
# 0x4038B7C0    3325816    2938  *Packet Header*
# Displayed first 2048 Allocator PCs only
# 0x4423D158     138992    1810  Init
# 0x41FD28F8     116992    1327  Init
# 0x41A873F4      95204    1226  Parser Linkage
# 0x4380A8C4     234076    1202  Init
# 0x43814134     120128     704  Init
# 0x41FD8754      61904     614  Init
# <snip>
#
# You will note that "Displayed first 2048 Allocator PCs only was also 
# included in the output, this is because the third column contained an
# integer.  This output is sorted by the third column which starts with 
# 2938 and decreases from there.
#

# sort_string { t_string index }
# Sort a string output from a command by first splitting it into list where
# each line of the string is a list element, then loop through the list 
# elements ignoring ones with too few columns and those that do not have
# integer values in the column being sorted.  A column is any word seperated
# by white space.  Once we have the list of lines to sort over, return a
# sorted list based on the index.  
proc sort_string { t_string index } {
    set t_list [split $t_string "\n"]
    set t {}
    foreach element $t_list {
        if {[llength $element] < $index} {
            # Ignore lines with not enough columns
            continue
        } elseif {![string is integer [lindex $element [expr $index - 1]]]} {
            # Ignore lines where the index column is not an integer
            continue
        } else {
            # Append this line to the end of the list
            lappend t $element
        }
    }
    # Sort the list in decreasing order using dictionary sorting based on the
    # index column.
    return [lsort -dictionary -decreasing -index [expr $index -1] $t]
}

# do_action { original_command command_tokens }
# This is the action proc that gets the command
# output, passes it to the sorting function and then
# prints the output to the stdout using puts.
proc do_action { original_command command_tokens } {

    # Open a vty to execute the command:
    if [catch {cli_open} result] {
        error $result $errorInfo
    } else {
        array set cli1 $result
    }

    # Go into enable mode
    if [catch {cli_exec $cli1(fd) "enable"} result] {
        error $result $errorInfo
    }

    # Execute the command:
    if [catch {cli_exec $cli1(fd) $original_command} result] {
        error $result $errorInfo
    } else {
        set command_output $result
    }

    # Close the vty
    if [catch {cli_close $cli1(fd) $cli1(tty_id)} result] {
        error $result $errorInfo
    }
    # Sort the output of the command using the sort_string procedure, this
    # returns a list of lines that have been sorted based on the column entered
    # join the list (the lines contain newlines so join with an empty string
    # and print the result.
    puts [join [sort_string $command_output [lindex $command_tokens 1]] ""]
}

# parse_remove_tokens_before_pipe { line }
# Remove all tokens after a pipe, this also removes the 
# pipe and the surrounding spaces which are required by
# the IOS parser.  This is used to add output modifiers
# to the IOS CLI.
#
# Example:
# Given :
#     $line: sh ev man pol reg | sort 1
# Execute:
#     parse_remove_expanded_tokens($line)
# Result:
#     Return string: tokens {sort 1} end_space 0
#
proc parse_remove_tokens_before_pipe { line } {

    set add_space 0

    if {[string index $line end] == " "} {
        set add_space 1
    }     

    set pipe_index [expr [string last " | " $line] + 3]
    if {$pipe_index > 2} {
        set newstring [string range $line $pipe_index end]
    } else {
        set newstring ""
    }

    return "tokens {$newstring} end_space $add_space"
}

# parse_remove_tokens_after_pipe { line }
# Remove all tokens before a pipe, this also removes the 
# pipe and the surrounding spaces which are required by
# the IOS parser.  This is used to add output modifiers
# to the IOS CLI.
#
# Example:
# Given :
#     $line: sh ver | sort 1
# Execute:
#     parse_remove_expanded_tokens($line)
# Result:
#     Return string: tokens {sh ver}
#
proc parse_remove_tokens_after_pipe { line } {

    set pipe_index [string last " | " $line]
    if {$pipe_index != -1 } {
        set newstring [string range $line 0 $pipe_index]
    } else {
        set newstring ""
    }

    return "tokens {$newstring}"
}

# insert_help_if_matched { <lots of args> }
# This started off as a simple procedure to insert help into
# the parser parse tree if the token matched what the user 
# entered, but it has evolved into a more complicated 
# procedure that does many things specific to this script.
# Not very portable.
proc insert_help_if_matched { type tty help index line token tokenhelp insert } {

    if {$index == 0} {
        set match [::cisco::lib::parse_check_token_match $token \
            [lindex $line $index ]]
    } elseif {$index == 1} {
        set match [regexp {[set token]} $line dummy]
        # Famous last words, "1000 should be enough for every one"
        set token "<0-1000>"
        if {[llength $line] > 1 && [lindex $line $index] <= 0 || [lindex $line $index] > 1000 } {
            set match 0
        }
    } else {
        set match 0
    }
    if {$insert == 1 && $match == 1} {
        if {$type == "short"} {
            ::cisco::lib::parse_insert_token_into_help_list \
                $tty $token $help
        } elseif {$type == "long"} {
            set newhelp "  $token\t$tokenhelp\n"
            append newhelp $help
            action_help_buffer_set tty $tty help $newhelp
        }
    }
 
    return $match
}

namespace import ::cisco::eem::*
namespace import ::cisco::lib::*

array set arr_einfo [event_reqinfo]
if {$_cerrno != 0} {
    set result [format "component=%s; subss err=%s; posix err=%s;\n%s" \
     $_cerr_sub_num $_cerr_sub_err $_cerr_posix_err $_cerr_str]
     error $result
}

array set arr_help [action_help_buffer_get tty $arr_einfo(tty)]
if {$_cerrno != 0} {
   set result [format "component=%s; subsys err=%s; posix err=%s;\n%s" \
       $_cerr_sub_num $_cerr_sub_err $_cerr_posix_err $_cerr_str]
   error $result
}

set newtoken "sort"
set newtoken_help "Sort by column (Added by EEM)"

array set line_aft_pipe \
     [parse_remove_tokens_before_pipe $arr_einfo(line)]
array set line_bef_pipe \
     [parse_remove_tokens_after_pipe $arr_einfo(line)]

if {$arr_einfo(key) == "questionmark"} {
    # Handle questionmark help.
    # No need to handle pipe since we are already in the pipe :)
    # Return if this is an ambiguous command (error code 1).
    if {$arr_einfo(error_code) == 1} {
        return;
    }
    if {$arr_einfo(error_code) == 0} {
        # No error this can happen in the following circumstances
        # show ver | ? <- Insert our token into the long help string
        # show ver | s? <- Insert our token into the short help string
        # show ver | s ? <- Ambiguous command with "section"
        if {$line_aft_pipe(end_space) == 0} {
            set match [::cisco::lib::parse_check_token_match $newtoken \
                   [lindex $arr_einfo(line) end]] 
            if {$match == 1} {
                # show ver | s? case
                ::cisco::lib::parse_insert_token_into_help_list \
                    $arr_einfo(tty) $newtoken $arr_help(help)
            }
        } else {
            if {[llength $line_aft_pipe(tokens)] == 0} {
                # show ver | ? case
                set newhelp "  $newtoken\t$newtoken_help\n"
                append newhelp $arr_help(help)
                action_help_buffer_set tty $arr_einfo(tty) help $newhelp
            } elseif {$line_aft_pipe(tokens) == "s "} {
                # show ver | s ? case
                set newhelp "% Ambiguous command:  \"$arr_einfo(line)\""
                action_help_buffer_set tty $arr_einfo(tty) help $newhelp
            }
        }
    } else {
        # If the error code is non-0 then the user entered a command
        # modifier that is not known by the system natively.
        # The user typed ? after some other string, need to see
        # if it matches the new token.
        if {$line_aft_pipe(end_space) == 0} {
            # No ending space, so do short form help
            set num_tokens [llength $line_aft_pipe(tokens)]
            if {$num_tokens == 1} {
                # Check if we match the sort token and insert sort into the help if we do.
                insert_help_if_matched "short" $arr_einfo(tty) $arr_help(help) 0 \
                    $line_aft_pipe(tokens) $newtoken "" 1
            } elseif {$num_tokens == 2} {
                # Check if we match the sort token and an integer
                if {[insert_help_if_matched "short" $arr_einfo(tty) $arr_help(help) 0 \
                    $line_aft_pipe(tokens) $newtoken "" 0 ] == 1} {
                    # sort matched, now check if the integer matched and if it does
                    # insert it into the help
                    insert_help_if_matched "short" $arr_einfo(tty) $arr_help(help) 1 \
                        $line_aft_pipe(tokens) {^[0-9]+$} "" 1
                }
            }
        } else {
            # Do the long form help.  This occurs if there is a space at the end of
            # the tokens.
            set num_tokens [llength $line_aft_pipe(tokens)]
            if {$num_tokens == 1} {
                # Check if we match the sort token and insert sort into the help if we do
                if {[insert_help_if_matched "long" $arr_einfo(tty) $arr_help(help) 0 \
                    $line_aft_pipe(tokens) $newtoken "" 0 ] == 1} {
                    # sort matched, now check if the integer matched and if it does
                    # insert it into the help
                    insert_help_if_matched "long" $arr_einfo(tty) $arr_help(help) 1 \
                        $line_aft_pipe(tokens) {^[0-9]+$} "Column number" 1
                }
            } elseif {$num_tokens == 2} {
                # Check if we match the sort token and an integer
                if {[insert_help_if_matched "long" $arr_einfo(tty) $arr_help(help) 0 \
                    $line_aft_pipe(tokens) $newtoken "" 0 ] == 1} {
                    # sort matched, now check if the integer matched and if it does add
                    # cr into the help since we have "sort <int> ".
                    if {[insert_help_if_matched "long" $arr_einfo(tty) $arr_help(help) 1 \
                        $line_aft_pipe(tokens) {^[0-9]+$} "<cr>" 0] == 1} {
                            action_help_buffer_set tty $arr_einfo(tty) help "  <cr>"
                    }
                }
            }              
        }
    }
}

if {$arr_einfo(key) == "tab"} {
    # See if last token typed is in correct location and can match keyword,
    # if yes then add keyword entry to help buffer. 
    set match [::cisco::lib::parse_check_token_match $newtoken \
             [lindex $line_aft_pipe(tokens) 0]]
    if {$match == 1} {
        # When the token is inserted into the help list the parser takes
        # care of command completion for us.
        ::cisco::lib::parse_insert_token_into_help_list $arr_einfo(tty) \
                                 $newtoken $arr_help(help)
    }     
}

if {$arr_einfo(key) == "enter"} {
    # Handle enter control character.
    # If we don't have at least two tokens or if we
    # have more than two, error out with the approriate error.
    if {[llength $line_aft_pipe(tokens)] < 2} {
        # We need at least "sort <int>"
        puts "% Incomplete command."
        exit 0
    } elseif {[llength $line_aft_pipe(tokens)] > 2} {
        # It can't be anything more than "sort <int>" and if it is,
        # it is an unknown command.
        exit 1
    }
    # Make sure we match at least part of the token before trying to
    # do the sort
    set match [::cisco::lib::parse_check_token_match $newtoken \
             [lindex $line_aft_pipe(tokens) 0]]
    if {$match == 1} {
        do_action $line_bef_pipe(tokens) $line_aft_pipe(tokens)
    } else {
        # We didn't match anything so either it is an invalid command
        # or the parser is supposed to handle it.
        exit 1
    }
}
